#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.zmi.trigger module

This module defines views used for tasks triggers management.
"""

from io import StringIO

from zope.interface import Interface

from pyams_form.ajax import ajax_form_config
from pyams_form.button import Buttons, handler
from pyams_form.field import Fields
from pyams_form.interfaces.form import IAJAXFormRenderer, IGroup
from pyams_layer.interfaces import IPyAMSLayer
from pyams_scheduler.interfaces import IScheduler, ITask, MANAGE_TASKS_PERMISSION
from pyams_scheduler.interfaces.task import ICronTask, ICronTaskScheduling, IDateTask, \
    IDateTaskScheduling, ILoopTask, ILoopTaskScheduling
from pyams_scheduler.task.zmi import TaskBaseFormMixin
from pyams_scheduler.zmi import SchedulerTasksTable
from pyams_skin.interfaces.viewlet import IHelpViewletManager
from pyams_skin.schema.button import SubmitButton
from pyams_skin.viewlet.help import AlertMessage
from pyams_table.interfaces import IColumn
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.traversing import get_parent
from pyams_viewlet.viewlet import EmptyViewlet, viewlet_config
from pyams_zmi.form import AdminModalEditForm, FormGroupChecker
from pyams_zmi.helper.event import get_json_table_row_refresh_callback
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.form import IModalEditFormButtons
from pyams_zmi.table import ActionColumn


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


#
# Task immediate run views
#

@adapter_config(name='run',
                required=(IScheduler, IAdminLayer, SchedulerTasksTable),
                provides=IColumn)
class SchedulerTaskRunColumn(ActionColumn):
    """Scheduler task run column"""

    href = 'run.html'
    icon_class = 'far fa-play-circle'
    hint = _("Run task")

    permission = MANAGE_TASKS_PERMISSION

    weight = 40


class ITaskRunButtons(IModalEditFormButtons):
    """Task run buttons"""

    debug = SubmitButton(name='debug',
                         title=_("Run in debug mode"))

    run = SubmitButton(name='run',
                       title=_("Run task"))


@ajax_form_config(name='run.html', context=ITask, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class TaskRunEditForm(TaskBaseFormMixin, AdminModalEditForm):
    """Task run edit form"""

    legend = _("Task execution")
    modal_class = 'modal-xl'
    has_border = True

    fields = Fields(Interface)
    buttons = Buttons(ITaskRunButtons).select('debug', 'run', 'close')

    def update_actions(self):
        super().update_actions()
        debug = self.actions.get('debug')
        if debug is not None:
            debug.add_class('btn-info mr-auto')

    @handler(ITaskRunButtons['debug'])
    def handle_debug(self, action):
        """Debug button handler"""
        report = StringIO()
        self.context.run(report)
        self.finished_state.update({
            'action': action,
            'changes': report.getvalue()
        })

    @handler(ITaskRunButtons['run'])
    def handle_apply(self, action):
        """Task run button handler"""
        self.context.launch()
        self.finished_state.update({
            'action': action,
            'changes': self.context
        })


@adapter_config(name='debug',
                required=(ITask, IAdminLayer, TaskRunEditForm),
                provides=IAJAXFormRenderer)
class TaskRunFormDebugActionRenderer(ContextRequestViewAdapter):
    """Task run form 'debug' action renderer"""

    def render(self, changes):
        """AJAX form renderer"""
        if not changes:
            return None
        return {
            'status': 'success',
            'message': self.request.localizer.translate(_("Task has been executed!")),
            'content': {
                'target': '#{}-debug-target'.format(self.view.id),
                'text': self.view.finished_state.get('changes', '-- NO OUTPUT --')
            },
            'closeForm': False
        }


@adapter_config(name='run',
                required=(ITask, IAdminLayer, TaskRunEditForm),
                provides=IAJAXFormRenderer)
class TaskRunFormRunActionRenderer(ContextRequestViewAdapter):
    """Task run form 'run' action renderer"""

    def render(self, changes):
        """AJAX form renderer"""
        if not changes:
            return None
        return {
            'status': 'success',
            'message': self.request.localizer.translate(_("The task has been planned for "
                                                          "immediate execution...<br />"
                                                          "Check task history to get execution "
                                                          "report."))
        }


@viewlet_config(name='task-run.help',
                context=ITask, layer=IAdminLayer, view=TaskRunEditForm,
                manager=IHelpViewletManager, weight=10)
class TaskRunFormHelp(AlertMessage):
    """Task run form help"""

    _message = _("You can run the task in normal mode or in debug mode.\n"
                 "In debug mode, the task is started in current application thread, and you will "
                 "get the output directly in this form.\n"
                 "In normal mode, the task is scheduled as usual, but for immediate execution.\n"
                 "Please note that in both modes, the scheduling mode is not taken into account, "
                 "including when the task is disabled!")


@viewlet_config(name='task-debug.target',
                context=ITask, layer=IAdminLayer, view=TaskRunEditForm,
                manager=IHelpViewletManager, weight=20)
class TaskRunFormDebugTarget(EmptyViewlet):
    """Task run form debug target"""

    def render(self):
        """Viewlet renderer"""
        return '<textarea id="{}-debug-target" ' \
               ' class="alert alert-secondary border-secondary text-monospace ' \
               '        w-100 height-300px p-2 hidden"></textarea>'.format(self.view.id)


#
# Task schedule views
#

@adapter_config(name='schedule',
                required=(IScheduler, IAdminLayer, SchedulerTasksTable),
                provides=IColumn)
class SchedulerTaskScheduleColumn(ActionColumn):
    """Scheduler task schedule column"""

    href = 'schedule.html'
    icon_class = 'fas fa-clock'
    hint = _("Schedule task")

    permission = MANAGE_TASKS_PERMISSION

    weight = 50


@ajax_form_config(name='schedule.html', context=ITask, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class TaskScheduleEditForm(TaskBaseFormMixin, AdminModalEditForm):
    """Task schedule edit form"""

    legend = _("Task schedule properties")
    fields = Fields(Interface)

    def apply_changes(self, data):
        changes = super().apply_changes(data)
        if changes:
            self.context.reset()
        return changes


@adapter_config(required=(ITask, IAdminLayer, TaskScheduleEditForm),
                provides=IAJAXFormRenderer)
class TaskEditFormAJAXRenderer(ContextRequestViewAdapter):
    """Task schedule edit form AJAX renderer"""

    def render(self, changes):
        """AJAX result renderer"""
        if not changes:
            return None
        scheduler = get_parent(self.context, IScheduler)
        return {
            'callbacks': [
                get_json_table_row_refresh_callback(scheduler, self.request,
                                                    SchedulerTasksTable, self.context)
            ]
        }


@adapter_config(name='cron-task-schedule.group',
                required=(ICronTask, IAdminLayer, TaskScheduleEditForm),
                provides=IGroup)
class CronTaskScheduleEditFormGroup(FormGroupChecker):
    """Cron-style task schedule edit form"""

    fields = Fields(ICronTaskScheduling)


@viewlet_config(name='cron-task-schedule.help',
                context=ICronTask, layer=IAdminLayer, view=CronTaskScheduleEditFormGroup,
                manager=IHelpViewletManager, weight=1)
class CronTaskScheduleEditFormHelp(AlertMessage):
    """Cron-style task schedule edit form help"""

    _message = _("You can enter values like in a \"classical\" crontab: you can set individual "
                 "values, ranges (using a dash as separator), steps (using */x), or multiple "
                 "values (using a comma as separator); steps are used to start a task, for "
                 "example, every five minutes, in the form \"*/5\".\n"
                 "For example: a value of \"3-6,8,10,12,14,16-19\" in the \"hours\" field would "
                 "mean that the task is to be run each hour from 3 to 6, then at 8, 10, 12 and "
                 "14, and finally each hour from 16 to 19; by setting \"*/2\" in the month days "
                 "field, the task would only run one day out of two.\n"
                 "An asterisk means that all possible values are selected (warning when setting "
                 "this value on seconds!!)")


@adapter_config(name='date-task-schedule.group',
                required=(IDateTask, IAdminLayer, TaskScheduleEditForm),
                provides=IGroup)
class DateTaskScheduleEditFormGroup(FormGroupChecker):
    """Date-style task schedule edit form"""

    fields = Fields(IDateTaskScheduling)


@adapter_config(name='loop-task-schedule.group',
                required=(ILoopTask, IAdminLayer, TaskScheduleEditForm),
                provides=IGroup)
class LoopTaskScheduleEditFormGroup(FormGroupChecker):
    """Loop-style task schedule edit form"""

    fields = Fields(ILoopTaskScheduling)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.task.handler base module

This module defines base directory handler class.
"""

__docformat__ = 'restructuredtext'

import os
from contextlib import contextmanager

from zope.interface import implementer

from pyams_scheduler.interfaces.task.sync import IDirectoryHandler
from pyams_utils.adapter import ContextAdapter


@implementer(IDirectoryHandler)
class BaseDirectoryHandler(ContextAdapter):
    """Base directory handler"""

    directory = None

    def connect(self):
        """Open connection"""

    def close(self):
        """Close handler connection"""

    def get_path(self, name):
        """Get full path for given name"""
        if name.startswith(os.path.sep):
            return name
        return os.path.join(self.directory, name)

    @contextmanager
    def chdir(self, dirname):
        """Change current directory"""
        old_directory = self.directory
        try:
            self.directory = self.get_path(dirname)
            yield
        finally:
            self.directory = old_directory

    @contextmanager
    def open(self, name, mode='r'):
        """Open file"""

    def mkdir(self, dirname, mode=0o500):
        """Create new directory"""

    def listdir(self):
        """List current directory content"""

    def stat(self, name):
        """Get stats about given file name"""

    def set_time(self, name, atime, mtime):
        """Set atime and mtime of given file name"""

    def set_mode(self, name, mode=0x400):
        """Set mode of given file name"""

    def unlink(self, name):
        """Delete given file name"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.zmi.task.ssh module

This modules defines SSH task management interface.
"""

from zope.interface import Interface, implementer

from pyams_form.ajax import ajax_form_config
from pyams_form.browser.object import ObjectWidget
from pyams_form.field import Fields
from pyams_form.group import GroupManager
from pyams_form.interfaces import IObjectFactory
from pyams_form.interfaces.form import IForm, IInnerTabForm
from pyams_form.interfaces.widget import IObjectWidget
from pyams_form.subform import InnerAddForm, InnerEditForm
from pyams_form.widget import FieldWidget
from pyams_layer.interfaces import IFormLayer, IPyAMSLayer
from pyams_scheduler.interfaces import IScheduler, MANAGE_TASKS_PERMISSION
from pyams_scheduler.interfaces.task.ssh import ISSHCallTaskInfo, ISSHCallerTask, \
    ISSHConnectionInfo
from pyams_scheduler.task.ssh import SSHCallerTask, SSHConnectionInfo
from pyams_scheduler.zmi import SchedulerTasksTable
from pyams_scheduler.task.zmi import BaseTaskAddForm, BaseTaskEditForm
from pyams_skin.viewlet.menu import MenuItem
from pyams_utils.adapter import adapter_config
from pyams_utils.factory import get_interface_name
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.viewlet import IContextAddingsViewletManager


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


#
# SSH connection object widget
#

@adapter_config(name=get_interface_name(ISSHConnectionInfo),
                required=(Interface, IFormLayer, IForm, IObjectWidget),
                provides=IObjectFactory)
def ssh_connection_factory(*args, **kwargs):  # pylint: disable=unused-argument
    """SSH connection object factory"""
    return SSHConnectionInfo


class SSHConnectionWidget(ObjectWidget):
    """SSH connection widget"""

    def update_widgets(self, set_errors=True):
        super().update_widgets(set_errors)
        password = self.widgets.get('password')  # pylint: disable=no-member
        if password is not None:
            password.autocomplete = 'new-password'


def SSHConnectionFieldWidget(field, request):  # pylint: disable=invalid-name
    """SSH connection field widget"""
    return FieldWidget(field, SSHConnectionWidget(request))


#
# Base SSH task form
#

class ISSHTaskForm(IForm):
    """SSH caller forms common interface"""


@implementer(ISSHTaskForm)
class SSHTaskFormInfo(GroupManager):
    """SSH task form info"""

    title = _("Command settings")
    fields = Fields(ISSHCallTaskInfo)
    fields['connection'].widget_factory = SSHConnectionFieldWidget


#
# SSH task add form
#

@viewlet_config(name='add-ssh-task.menu',
                context=IScheduler, layer=IAdminLayer, view=SchedulerTasksTable,
                manager=IContextAddingsViewletManager, weight=10,
                permission=MANAGE_TASKS_PERMISSION)
class SSHTaskAddMenu(MenuItem):
    """SSH task add menu"""

    label = _("Add local/remote command launcher...")
    href = 'add-ssh-task.html'
    modal_target = True


@ajax_form_config(name='add-ssh-task.html', context=IScheduler, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class SSHTaskAddForm(BaseTaskAddForm):
    """SSH task add form"""

    content_factory = ISSHCallerTask
    content_label = SSHCallerTask.label


@adapter_config(name='ssh-task-info.form',
                required=(IScheduler, IAdminLayer, SSHTaskAddForm),
                provides=IInnerTabForm)
class SSHTaskAddFormInfo(SSHTaskFormInfo, InnerAddForm):
    """SSH task add form info"""


#
# SSH task edit form
#

@ajax_form_config(name='properties.html', context=ISSHCallerTask, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class SSHTaskEditForm(BaseTaskEditForm):
    """SSH task edit form"""


@adapter_config(name='ssh-task-info.form',
                required=(ISSHCallerTask, IAdminLayer, SSHTaskEditForm),
                provides=IInnerTabForm)
class SSHTaskEditFormInfo(SSHTaskFormInfo, InnerEditForm):
    """SSH task edit form info"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.zmi.scheduler module

This module defines main scheduler tasks view.
"""

from zope.interface import Interface

from pyams_form.ajax import ajax_form_config
from pyams_form.field import Fields
from pyams_layer.interfaces import IPyAMSLayer
from pyams_scheduler.interfaces import IScheduler, MANAGE_SCHEDULER_PERMISSION, \
    TASKS_SCHEDULER_LABEL
from pyams_site.interfaces import ISiteRoot
from pyams_skin.interfaces.viewlet import IBreadcrumbItem
from pyams_utils.adapter import adapter_config
from pyams_utils.registry import get_utility, query_utility
from pyams_utils.url import absolute_url
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminEditForm
from pyams_zmi.interfaces import IAdminLayer, IObjectLabel
from pyams_zmi.interfaces.viewlet import IControlPanelMenu, ISiteManagementMenu
from pyams_zmi.zmi.viewlet.breadcrumb import AdminLayerBreadcrumbItem
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


@adapter_config(required=(IScheduler, IPyAMSLayer, Interface),
                provides=IObjectLabel)
def scheduler_label(context, request, view):  # pylint: disable=unused-argument
    """Scheduler label"""
    return request.localizer.translate(TASKS_SCHEDULER_LABEL)


@adapter_config(required=(IScheduler, IAdminLayer, Interface),
                provides=IBreadcrumbItem)
class SchedulerBreadcrumbItem(AdminLayerBreadcrumbItem):
    """Scheduler breadcrumb item"""

    label = TASKS_SCHEDULER_LABEL


@viewlet_config(name='scheduler.menu',
                context=ISiteRoot, layer=IAdminLayer,
                manager=IControlPanelMenu, weight=25,
                permission=MANAGE_SCHEDULER_PERMISSION)
class SchedulerMenu(NavigationMenuItem):
    """Scheduler root menu"""

    label = TASKS_SCHEDULER_LABEL
    icon_class = 'fa fa-clock'

    def __new__(cls, context, request, view, manager):  # pylint: disable=unused-argument
        scheduler = query_utility(IScheduler)
        if (scheduler is None) or not scheduler.show_home_menu:
            return None
        return NavigationMenuItem.__new__(cls)

    def get_href(self):
        """Menu URL getter"""
        scheduler = get_utility(IScheduler)
        return absolute_url(scheduler, self.request, 'admin')


@viewlet_config(name='configuration.menu',
                context=IScheduler, layer=IAdminLayer,
                manager=ISiteManagementMenu, weight=20,
                permission=MANAGE_SCHEDULER_PERMISSION)
class SchedulerConfigurationMenu(NavigationMenuItem):
    """Scheduler configuration menu"""

    label = _("Configuration")
    icon_class = 'fas fa-sliders-h'
    href = '#configuration.html'


@ajax_form_config(name='configuration.html', context=IScheduler, layer=IPyAMSLayer,
                  permission=MANAGE_SCHEDULER_PERMISSION)
class SchedulerConfigurationEditForm(AdminEditForm):
    """Scheduler configuration edit form"""

    title = TASKS_SCHEDULER_LABEL
    legend = _("Scheduler configuration")

    fields = Fields(IScheduler).select('zodb_name', 'report_mailer', 'report_source',
                                       'notified_host', 'show_home_menu')



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

__docformat__ = 'restructuredtext'

from pyams_form.interfaces import DISPLAY_MODE, INPUT_MODE
from pyams_form.interfaces.form import IForm
from pyams_form.interfaces.widget import IWidget
from pyams_form.template import widget_template_config
from pyams_layer.interfaces import IPyAMSLayer


@widget_template_config(mode=INPUT_MODE,
                        template='templates/directory-host-input.pt', layer=IPyAMSLayer)
@widget_template_config(mode=DISPLAY_MODE,
                        template='templates/directory-host-display.pt', layer=IPyAMSLayer)
class IDirectoryHandlerHostWidget(IWidget):
    """Directory handler host widget interface"""


class ITaskForm(IForm):
    """Base task form interface"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.zmi.task base module

This module defines base tasks management forms.
"""

from zope.copy import copy
from zope.interface import Interface
from zope.intid import IIntIds

from pyams_form.ajax import ajax_form_config
from pyams_form.field import Fields
from pyams_form.interfaces.form import IAJAXFormRenderer, IInnerTabForm
from pyams_form.subform import InnerAddForm, InnerEditForm
from pyams_layer.interfaces import IPyAMSLayer
from pyams_scheduler.interfaces import IScheduler, ITask, MANAGE_TASKS_PERMISSION
from pyams_scheduler.interfaces.task import ITaskInfo
from pyams_scheduler.zmi import SchedulerTasksTable
from pyams_skin.viewlet.help import AlertMessage
from pyams_table.interfaces import IColumn
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.registry import get_utility, query_utility
from pyams_utils.traversing import get_parent
from pyams_zmi.form import AdminModalAddForm, AdminModalEditForm
from pyams_zmi.helper.event import get_json_table_row_add_callback, \
    get_json_table_row_refresh_callback
from pyams_zmi.interfaces import IAdminLayer, IObjectLabel
from pyams_zmi.interfaces.table import ITableElementEditor
from pyams_zmi.table import ActionColumn, TableElementEditor
from pyams_zmi.utils import get_object_label


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


#
# Task history mixin inner form
#

class TaskHistoryHelpMessage(AlertMessage):
    """Task history help message"""

    css_class = 'mb-1 p-2'
    _message = _("You can limit history conservation to a duration or to a number of iterations. "
                 "If both are specified, the first encountered limit will take precedence.")


#
# Base task add form
#

class BaseTaskAddForm(AdminModalAddForm):  # pylint: disable=abstract-method
    """Base task add form"""

    @property
    def title(self):
        """Title getter"""
        translate = self.request.localizer.translate
        return '<small>{}</small><br />{}'.format(
            get_object_label(self.context, self.request, self),
            translate(_("New task: {}")).format(translate(self.content_label)))

    legend = _("New task properties")
    content_label = '--'

    fields = Fields(ITaskInfo).select('name', 'schedule_mode')

    def add(self, obj):
        intids = get_utility(IIntIds)
        self.context[hex(intids.register(obj))[2:]] = obj


@adapter_config(name='base-task-info',
                required=(IScheduler, IAdminLayer, BaseTaskAddForm),
                provides=IInnerTabForm)
class BaseTaskAddFormInfo(InnerAddForm):
    """Base task add form general information tab"""

    title = _("Task reports")

    fields = Fields(ITaskInfo).select('keep_empty_reports', 'history_duration',
                                      'history_length')
    weight = 100

    def update_widgets(self, prefix=None):
        """Widgets update"""
        super().update_widgets(prefix)
        widget = self.widgets.get('history_duration')
        if widget is not None:
            widget.prefix = TaskHistoryHelpMessage(self.context, self.request, self, None)


@adapter_config(required=(IScheduler, IAdminLayer, BaseTaskAddForm),
                provides=IAJAXFormRenderer)
class TaskAddFormAJAXRenderer(ContextRequestViewAdapter):
    """Base task add form AJAX renderer"""

    def render(self, changes):
        """AJAX result renderer"""
        if not changes:
            return None
        return {
            'callbacks': [
                get_json_table_row_add_callback(self.context, self.request,
                                                SchedulerTasksTable, changes)
            ]
        }


#
# Base task edit form
#

@adapter_config(required=(ITask, IPyAMSLayer, Interface),
                provides=IObjectLabel)
def task_label(context, request, view):  # pylint: disable=unused-argument
    """Task table element name factory"""
    return context.name


@adapter_config(required=(ITask, IAdminLayer, Interface),
                provides=ITableElementEditor)
class TaskTableElementEditor(TableElementEditor):
    """Task table element editor"""


class TaskBaseFormMixin:
    """Task base edit form mixin"""

    @property
    def title(self):
        """Title getter"""
        translate = self.request.localizer.translate
        scheduler = query_utility(IScheduler)
        task = get_parent(self.context, ITask)
        return '<small>{}</small><br />{}'.format(
            get_object_label(scheduler, self.request, self),
            translate(_("Task: {}")).format(task.name))


class BaseTaskEditForm(TaskBaseFormMixin, AdminModalEditForm):
    """Base task edit form"""

    legend = _("Task properties")

    fields = Fields(ITaskInfo).select('name', 'schedule_mode')

    def apply_changes(self, data):
        changes = super().apply_changes(data)
        if changes:
            self.context.reset()
        return changes


@adapter_config(name='base-task-info',
                required=(ITask, IAdminLayer, BaseTaskEditForm),
                provides=IInnerTabForm)
class BaseTaskEditFormInfo(InnerEditForm):
    """Base task edit form general information tab"""

    title = _("Task reports")

    fields = Fields(ITaskInfo).select('keep_empty_reports', 'history_duration',
                                      'history_length')
    weight = 100

    def update_widgets(self, prefix=None):
        """Widgets update"""
        super().update_widgets(prefix)
        widget = self.widgets.get('history_duration')
        if widget is not None:
            widget.prefix = TaskHistoryHelpMessage(self.context, self.request, self, None)


@adapter_config(required=(ITask, IAdminLayer, BaseTaskEditForm),
                provides=IAJAXFormRenderer)
class TaskEditFormAJAXRenderer(ContextRequestViewAdapter):
    """Task edit form AJAX renderer"""

    def render(self, changes):
        """AJAX result renderer"""
        if not changes:
            return None
        scheduler = get_parent(self.context, IScheduler)
        return {
            'callbacks': [
                get_json_table_row_refresh_callback(scheduler, self.request,
                                                    SchedulerTasksTable, self.context)
            ]
        }


#
# Task clone form
#

@adapter_config(name='clone',
                required=(IScheduler, IAdminLayer, SchedulerTasksTable),
                provides=IColumn)
class TaskCloneColumn(ActionColumn):
    """Task clone column"""

    hint = _("Clone task")
    icon_class = 'far fa-clone'

    href = 'clone-task.html'

    weight = 100


@ajax_form_config(name='clone-task.html', context=ITask,
                  layer=IPyAMSLayer, permission=MANAGE_TASKS_PERMISSION)
class TaskCloneForm(TaskBaseFormMixin, AdminModalAddForm):
    """Task clone form"""

    legend = _("Clone task")

    fields = Fields(ITask).select('name')

    def create(self, data):
        return copy(self.context)

    def add(self, obj):
        intids = get_utility(IIntIds)
        self.context.__parent__[hex(intids.register(obj))[2:]] = obj


@adapter_config(required=(ITask, IAdminLayer, TaskCloneForm),
                provides=IAJAXFormRenderer)
class TaskCloneFormAJAXRenderer(ContextRequestViewAdapter):
    """Task clone form AJAX renderer"""

    def render(self, changes):
        """AJAX result renderer"""
        if not changes:
            return None
        scheduler = get_parent(self.context, IScheduler)
        return {
            'callbacks': [
                get_json_table_row_add_callback(scheduler, self.request,
                                                SchedulerTasksTable, changes)
            ]
        }



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.task module

This module defines base tasks classes.
"""

import logging
import traceback
from datetime import datetime
from io import StringIO

from persistent import Persistent
from pyramid.config import Configurator
from pyramid.events import subscriber
from pyramid.threadlocal import RequestContext, manager
from transaction.interfaces import ITransactionManager
from zope.component import queryUtility
from zope.component.interfaces import ISite
from zope.container.contained import Contained
from zope.container.folder import Folder
from zope.copy.interfaces import ICopyHook, ResumeCopy
from zope.interface import alsoProvides, implementer, noLongerProvides
from zope.intid import IIntIds
from zope.lifecycleevent import IObjectAddedEvent, IObjectModifiedEvent, IObjectRemovedEvent
from zope.location import locate
from zope.schema.fieldproperty import FieldProperty


try:
    from pyams_chat.message import ChatMessage
except ImportError:
    ChatMessage = None

from pyams_scheduler.interfaces import AfterRunJobEvent, BeforeRunJobEvent, IScheduler, ITask, \
    ITaskHistory, MANAGE_TASKS_PERMISSION, SCHEDULER_AUTH_KEY, SCHEDULER_HANDLER_KEY, \
    SCHEDULER_MANAGER_ROLE, SCHEDULER_NAME, TASKS_MANAGER_ROLE
from pyams_scheduler.interfaces.task import FailedTaskRunException, ITaskHistoryContainer, \
    ITaskInfo, ITaskNotificationContainer, ITaskSchedulingMode, TASK_STATUS_CLASS, \
    TASK_STATUS_EMPTY, TASK_STATUS_ERROR, TASK_STATUS_FAIL, TASK_STATUS_NONE, TASK_STATUS_OK
from pyams_security.interfaces import IProtectedObject, IViewContextPermissionChecker
from pyams_security.interfaces.names import ADMIN_USER_ID, INTERNAL_USER_ID, SYSTEM_ADMIN_ROLE
from pyams_site.interfaces import PYAMS_APPLICATION_DEFAULT_NAME, PYAMS_APPLICATION_SETTINGS_KEY
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.date import get_duration
from pyams_utils.interfaces.transaction import ITransactionClient
from pyams_utils.registry import get_pyramid_registry, get_utility, query_utility, \
    set_local_registry
from pyams_utils.request import check_request
from pyams_utils.timezone import tztime
from pyams_utils.transaction import COMMITTED_STATUS, TransactionClient, transactional
from pyams_utils.traversing import get_parent
from pyams_utils.zodb import ZODBConnection
from pyams_zmq.socket import zmq_response, zmq_socket


__docformat__ = 'restructuredtext'

from pyams_scheduler import _


LOGGER = logging.getLogger('PyAMS (scheduler)')


@implementer(ITaskHistory)
class TaskHistoryItem(Persistent, Contained):
    """Task history item"""

    status = FieldProperty(ITaskHistory['status'])
    date = FieldProperty(ITaskHistory['date'])
    duration = FieldProperty(ITaskHistory['duration'])
    report = FieldProperty(ITaskHistory['report'])

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


@implementer(ITaskHistoryContainer)
class TaskHistoryContainer(Folder):
    """Task history container"""

    def check_history(self, duration, length):
        """Check history container contents"""
        if duration:
            now = tztime(datetime.utcnow())
            for key in list(self.keys()):
                if (now - tztime(self[key].date)).days > duration:
                    del self[key]
        if length and (len(self) > length):
            keys = sorted(self.keys(), reverse=True)[:length]
            for key in list(self.keys()):
                if key not in keys:
                    del self[key]


class TaskHandler(TransactionClient):
    """Generic task handler"""

    @transactional
    def execute(self, task, action, job_id, registry=None):
        """Execute scheduler action"""
        scheduler = query_utility(IScheduler)
        if scheduler is not None:
            registry = registry if registry is not None else get_pyramid_registry()
            handler = registry.settings.get(SCHEDULER_HANDLER_KEY, False)
            if handler:
                zmq_settings = {
                    'zodb_name': scheduler.zodb_name,
                    'task_name': task.__name__,
                    'job_id': job_id
                }
                LOGGER.debug(f"Starting '{action}' on task {task.name} with {zmq_settings!r}")
                socket = zmq_socket(handler, auth=registry.settings.get(SCHEDULER_AUTH_KEY))
                socket.send_json([action, zmq_settings])
                zmq_response(socket)


@implementer(ITask, ITransactionClient)
class Task(Persistent, Contained):
    """Task definition persistent class"""

    label = None
    icon_class = None

    name = FieldProperty(ITask['name'])
    _schedule_mode = FieldProperty(ITask['schedule_mode'])
    keep_empty_reports = FieldProperty(ITask['keep_empty_reports'])
    _history_duration = FieldProperty(ITask['history_duration'])
    _history_length = FieldProperty(ITask['history_length'])

    settings_view_name = FieldProperty(ITask['settings_view_name'])
    principal_id = None

    _internal_id = None

    def __init__(self):
        history = self.history = TaskHistoryContainer()
        locate(history, self, '++history++')

    @property
    def schedule_mode(self):
        """Scheduler mode getter"""
        return self._schedule_mode

    @schedule_mode.setter
    def schedule_mode(self, value):
        """Scheduler mode setter"""
        if self._schedule_mode is not None:
            mode = query_utility(ITaskSchedulingMode, name=self._schedule_mode)
            if (mode is not None) and mode.marker_interface.providedBy(self):
                noLongerProvides(self, mode.marker_interface)
        self._schedule_mode = value
        if value:
            mode = get_utility(ITaskSchedulingMode, name=value)
            alsoProvides(self, mode.marker_interface)
            mode.schema(self).active = False
            if self.__parent__ is not None:
                self.reset()

    @property
    def history_duration(self):
        """History duration getter"""
        return self._history_duration

    @history_duration.setter
    def history_duration(self, value):
        """History duration setter"""
        self._history_duration = value

    @property
    def history_length(self):
        """History length getter"""
        return self._history_length

    @history_length.setter
    def history_length(self, value):
        """History length setter"""
        self._history_length = value

    def check_history(self):
        """Check history container for old contents"""
        self.history.check_history(self.history_duration, self.history_length)

    @property
    def internal_id(self):
        """Task internal ID getter"""
        if self._internal_id is None:
            site = get_parent(self, ISite)
            sm = site.getSiteManager()  # pylint: disable=invalid-name,too-many-function-args,assignment-from-no-return
            intids = sm.queryUtility(IIntIds)
            if intids is not None:
                self._internal_id = intids.register(self)
        return self._internal_id

    def get_trigger(self):
        """Task trigger getter"""
        mode = queryUtility(ITaskSchedulingMode, self.schedule_mode)
        if mode is None:
            return None
        return mode.get_trigger(self)

    def get_scheduling_info(self):
        """Task scheduling info getter"""
        mode = queryUtility(ITaskSchedulingMode, self.schedule_mode)
        if mode is None:
            return None
        return mode.schema(self, None)

    def reset(self):
        """Task reset launcher"""
        handler = TaskHandler()
        handler.execute(self, 'reset_task', self.internal_id)

    def launch(self):
        """Task immediate launcher"""
        handler = TaskHandler()
        handler.execute(self, 'run_task', self.internal_id)

    def __call__(self, *args, **kwargs):
        report = StringIO()
        return self._run(report, **kwargs)

    def is_runnable(self):
        """Task runnable state getter"""
        mode = queryUtility(ITaskSchedulingMode, self.schedule_mode)
        if mode is None:
            return False
        info = mode.schema(self, None)
        if info is None:
            return False
        return info.active

    def _run(self, report, **kwargs):  # pylint: disable=too-many-locals
        """Task execution wrapper"""
        status = TASK_STATUS_NONE
        result = None
        # initialize ZCA hook
        registry = kwargs.pop('registry')
        manager.push({'registry': registry, 'request': None})
        config = Configurator(registry=registry)
        config.hook_zca()
        # open ZODB connection
        zodb_connection = ZODBConnection(name=kwargs.pop('zodb_name', ''))
        with zodb_connection as root:
            try:
                application_name = registry.settings.get(PYAMS_APPLICATION_SETTINGS_KEY,
                                                         PYAMS_APPLICATION_DEFAULT_NAME)
                application = root.get(application_name)
                sm = application.getSiteManager()  # pylint: disable=invalid-name
                scheduler_util = sm.get(SCHEDULER_NAME)
                task = scheduler_util.get(self.__name__)
                if task is not None:
                    set_local_registry(sm)
                    request = check_request(base_url=scheduler_util.notified_host,
                                            registry=registry, principal_id=self.principal_id)
                    request.root = application
                    with RequestContext(request):
                        if not (kwargs.get('run_immediate') or task.is_runnable()):
                            LOGGER.debug("Skipping inactive task {0}".format(task.name))
                            return status, result
                        translate = request.localizer.translate
                        tm = ITransactionManager(task)  # pylint: disable=invalid-name
                        for attempt in tm.attempts():
                            with attempt as t:  # pylint: disable=invalid-name
                                status = TASK_STATUS_NONE
                                start_date = datetime.utcnow()
                                duration = 0.
                                try:
                                    registry.notify(BeforeRunJobEvent(task))
                                    (status, result) = task.run(report, **kwargs)
                                    if status == TASK_STATUS_FAIL:
                                        raise FailedTaskRunException
                                    end_date = datetime.utcnow()
                                    duration = (end_date - start_date).total_seconds()
                                    report.write('\n\nTask duration: {0}'.format(
                                        get_duration(start_date, request=request)))
                                    if (ChatMessage is not None) and scheduler_util.notified_host:
                                        if status == TASK_STATUS_ERROR:
                                            message_text = translate(_("Task '{}' was executed "
                                                                       "with error")) \
                                                .format(task.name)
                                        else:
                                            message_text = translate(_("Task '{}' was executed "
                                                                       "without error")) \
                                                .format(task.name)
                                        message = ChatMessage(
                                            request=request,
                                            host=scheduler_util.notified_host,
                                            action='notify',
                                            category='scheduler.run',
                                            status=TASK_STATUS_CLASS.get(status, 'info'),
                                            source=INTERNAL_USER_ID,
                                            title=translate(_("Task execution")),
                                            message=message_text,
                                            url='/'.join(('', '++etc++site',
                                                          scheduler_util.__name__, 'admin'))
                                        )
                                        message.send()
                                except FailedTaskRunException:  # pylint: disable=bare-except
                                    # pylint: disable=protected-access
                                    task._log_exception(report,
                                                        "An error occurred during execution of "
                                                        "task '{}'".format(task.name))
                                    if (ChatMessage is not None) and scheduler_util.notified_host:
                                        message = ChatMessage(
                                            request=request,
                                            host=scheduler_util.notified_host,
                                            action='notify',
                                            category='scheduler.run',
                                            status=TASK_STATUS_CLASS.get(status, 'danger'),
                                            source=INTERNAL_USER_ID,
                                            title=translate(_("Task execution")),
                                            message=translate(_("An error occurred during "
                                                                "execution of task '{}'"
                                                                "")).format(task.name),
                                            url='/'.join(('', '++etc++site',
                                                          scheduler_util.__name__, 'admin'))
                                        )
                                        message.send()
                                registry.notify(AfterRunJobEvent(task, status, result))
                                task.store_report(report, status, start_date, duration)
                                task.send_report(report, status, registry)
                            if t.status == COMMITTED_STATUS:
                                break
            except:  # pylint: disable=bare-except
                self._log_exception(None, "Can't execute scheduled job {0}".format(self.name))
            tm = ITransactionManager(self, None)  # pylint: disable=invalid-name
            if tm is not None:
                tm.abort()
        return status, result

    def run(self, report, **kwargs):  # pylint: disable=no-self-use
        """Task run implementation

        May result a tuple containing a status code and a result (which can be empty).
        """
        raise NotImplementedError("The 'run' method must be implemented by Task subclasses!")

    @staticmethod
    def _log_report(report, message, add_timestamp=True, level=logging.INFO):
        """Execution log report"""
        if isinstance(message, bytes):
            message = message.decode()
        if add_timestamp:
            message = '{0} - {1}'.format(tztime(datetime.utcnow()).strftime('%c'), message)
        if report is not None:
            report.write(message + '\n')
        LOGGER.log(level, message)

    @staticmethod
    def _log_exception(report, message=None):
        """Exception log report"""
        if isinstance(message, bytes):
            message = message.decode()
        message = '{0} - {1}'.format(tztime(datetime.utcnow()).strftime('%c'),
                                     message or 'An error occurred') + '\n\n'
        if report is not None:
            report.write(message)
            report.write(traceback.format_exc() + '\n')
        LOGGER.exception(message)

    def store_report(self, report, status, start_date, duration):
        """Execution report store"""
        if (status in (TASK_STATUS_NONE, TASK_STATUS_EMPTY)) and \
                not self.keep_empty_reports:
            return
        item = TaskHistoryItem(status=str(status),
                               date=start_date,
                               duration=duration,
                               report=report.getvalue())
        self.history[item.date.isoformat()] = item
        self.check_history()

    def send_report(self, report, status, registry):
        """Execution report messaging"""
        notifications = ITaskNotificationContainer(self)
        for target in notifications.get_enabled_items():
            handler = target.get_handler()
            if handler is None:
                continue
            if ((status in (TASK_STATUS_NONE, TASK_STATUS_EMPTY)) and
                not target.send_empty_reports) or \
                    ((status == TASK_STATUS_OK) and target.report_errors_only):
                return

            handler.send_report(self, report, status, target, registry)


@subscriber(IObjectAddedEvent, context_selector=ITask)
def handle_new_task(event):
    """Handle new task"""
    event.object.reset()


@subscriber(IObjectModifiedEvent, context_selector=ITask)
def handle_modified_task(event):
    """Handle modified task"""
    for changes in event.descriptions:
        if (changes.interface == ITaskInfo) and \
                (('history_duration' in changes.attributes) or
                 ('history_length' in changes.attributes)):
            event.object.check_history()
            break


@subscriber(IObjectRemovedEvent, context_selector=ITask)
def handle_removed_task(event):
    """Handle removed task"""
    request = check_request()
    if request.registry is not None:
        handler = request.registry.settings.get(SCHEDULER_HANDLER_KEY, False)
        if handler:
            task = event.object
            scheduler_util = query_utility(IScheduler)
            zmq_settings = {
                'zodb_name': scheduler_util.zodb_name,
                'task_name': task.__name__,
                'job_id': task.internal_id
            }
            LOGGER.debug("Removing task {0.name} with {1!r}".format(task, zmq_settings))
            socket = zmq_socket(handler, auth=request.registry.settings.get(SCHEDULER_AUTH_KEY))
            socket.send_json(['remove_task', zmq_settings])
            zmq_response(socket)


@adapter_config(required=ITask,
                provides=ICopyHook)
class TaskCopyHook(ContextAdapter):
    """Task copy hook"""

    def __call__(self, toplevel, register):
        register(self._copy_history)
        raise ResumeCopy

    def _copy_history(self, translate):
        task = translate(self.context)
        task._internal_id = None  # pylint: disable=protected-access
        # create empty history
        history = task.history = TaskHistoryContainer()
        locate(history, task, '++history++')
        # disable task
        scheduling_mode = task.get_scheduling_info()
        scheduling_mode.active = False


@adapter_config(required=ITask,
                provides=IViewContextPermissionChecker)
class TaskPermissionChecker(ContextAdapter):
    """Task permission checker"""

    edit_permission = MANAGE_TASKS_PERMISSION


if ChatMessage is not None:

    from pyams_chat.interfaces import IChatMessage, IChatMessageHandler

    @adapter_config(name='scheduler.run',
                    required=IChatMessage,
                    provides=IChatMessageHandler)
    class SchedulerTaskRunMessageHandler(ContextAdapter):
        """Scheduler task run message handler"""

        def get_target(self):
            """Chat message targets getter"""
            principals = {ADMIN_USER_ID}
            root = self.context.request.root
            protection = IProtectedObject(root, None)
            if protection is not None:
                principals |= protection.get_principals(SYSTEM_ADMIN_ROLE)
            scheduler = get_utility(IScheduler)
            protection = IProtectedObject(scheduler, None)
            if protection is not None:
                principals |= protection.get_principals(SCHEDULER_MANAGER_ROLE)
                principals |= protection.get_principals(TASKS_MANAGER_ROLE)
            return {
                'principals': tuple(principals)
            }



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.task.handler.local module

"""

__docformat__ = 'restructuredtext'

import os
from contextlib import contextmanager

from pyams_scheduler.interfaces.task.sync import IDirectoryHandler, IDirectoryInfo
from pyams_scheduler.task.handler import BaseDirectoryHandler
from pyams_utils.adapter import adapter_config


@adapter_config(name='local',
                required=IDirectoryInfo,
                provides=IDirectoryHandler)
class LocalDirectoryHandler(BaseDirectoryHandler):
    """Local directory handler"""

    @contextmanager
    def open(self, name, mode='r'):
        file = open(self.get_path(name), mode)
        try:
            yield file
        finally:
            file.close()

    def mkdir(self, dirname, mode=0o500):
        os.mkdir(self.get_path(dirname), mode)

    def listdir(self):
        yield from sorted(os.listdir(self.directory))

    def stat(self, name):
        return os.stat(self.get_path(name))

    def set_time(self, name, atime, mtime):
        os.utime(self.get_path(name), times=(atime, mtime))

    def set_mode(self, name, mode=0x400):
        os.chmod(self.get_path(name), mode)

    def unlink(self, name):
        os.unlink(self.get_path(name))



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.task.sync module

This module provides main directory synchronization task components.
"""

import fnmatch
import io
import socket
import stat
import sys
import traceback
from contextlib import contextmanager

from persistent import Persistent
from zope.component import queryAdapter
from zope.schema.fieldproperty import FieldProperty

from pyams_scheduler.interfaces.task import TASK_STATUS_EMPTY, TASK_STATUS_ERROR, \
    TASK_STATUS_FAIL, TASK_STATUS_OK
from pyams_scheduler.interfaces.task.sync import DirectorySyncError, IDirectoryHandler, \
    IDirectoryInfo, IDirectorySyncTask
from pyams_scheduler.task import Task
from pyams_utils.factory import factory_config
from pyams_utils.size import get_human_size


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


@factory_config(IDirectoryInfo)
class DirectoryInfo(Persistent):
    """Directory info"""

    host = FieldProperty(IDirectoryInfo['host'])
    port = FieldProperty(IDirectoryInfo['port'])
    auto_add_host_key = FieldProperty(IDirectoryInfo['auto_add_host_key'])
    username = FieldProperty(IDirectoryInfo['username'])
    private_key = FieldProperty(IDirectoryInfo['private_key'])
    _password = FieldProperty(IDirectoryInfo['password'])
    directory = FieldProperty(IDirectoryInfo['directory'])

    def __init__(self, data=None):  # pylint: disable=unused-argument
        super().__init__()

    def __repr__(self):
        protocol, hostname = self.host or (None, None)
        if hostname:
            return '{}://{}@{}:{}{}'.format(protocol,
                                            self.username,
                                            hostname,
                                            self.port,
                                            self.directory)
        return '{}://{}{}'.format(protocol,
                                  socket.gethostname(),
                                  self.directory)

    @property
    def password(self):
        """Password getter"""
        return self._password

    @password.setter
    def password(self, value):
        """Password setter"""
        if value == '*****':
            return
        self._password = value

    @contextmanager
    def get_handler(self):
        """Get protocol handler"""
        protocol, _hostname = self.host or (None, None)
        if not protocol:
            return None
        handler = queryAdapter(self, IDirectoryHandler, name=protocol)
        if handler is not None:
            try:
                handler.directory = self.directory
                handler.connect()
                yield handler
            finally:
                if handler is not None:
                    handler.close()
        return None


@factory_config(IDirectorySyncTask)
class DirectorySyncTask(Task):
    """Directory synchronization task"""

    label = _("Directory synchronization")
    icon_class = 'fas fa-sync-alt'

    source = FieldProperty(IDirectorySyncTask['source'])
    filter = FieldProperty(IDirectorySyncTask['filter'])
    recursive = FieldProperty(IDirectorySyncTask['recursive'])
    follow_symlinks = FieldProperty(IDirectorySyncTask['follow_symlinks'])
    ignore_read_errors = FieldProperty(IDirectorySyncTask['ignore_read_errors'])
    use_datetime = FieldProperty(IDirectorySyncTask['use_datetime'])
    delete_source = FieldProperty(IDirectorySyncTask['delete_source'])
    ignore_delete_errors = FieldProperty(IDirectorySyncTask['ignore_delete_errors'])
    target = FieldProperty(IDirectorySyncTask['target'])
    ignore_write_errors = FieldProperty(IDirectorySyncTask['ignore_write_errors'])

    def run(self, report, **kwargs):  # pylint: disable=unused-argument,too-many-statements

        copied = {
            'files': [],  # list of copied files
            'size': 0
        }

        def copy_file(src, dst, name, src_stat):
            """Copy a single file from source to destination"""
            size = 0
            try:
                with src.open(name, 'rb') as src_file:
                    try:
                        with dst.open(name, 'wb') as dst_file:
                            data = src_file.read(io.DEFAULT_BUFFER_SIZE)
                            while data:
                                size += len(data)
                                dst_file.write(data)
                                data = src_file.read(io.DEFAULT_BUFFER_SIZE)
                        if self.use_datetime:
                            dst.set_time(name, src_stat.st_atime, src_stat.st_mtime)
                        dst.set_mode(name, src_stat.st_mode)
                        report.write('   ++ {}: {} ({})\n'.format(src.directory,
                                                                  name,
                                                                  get_human_size(size)))
                        copied['files'].append((src.directory, name))
                        copied['size'] += size
                    except PermissionError:
                        if not self.ignore_write_errors:
                            raise
                        report.write('   ** {}: {} (WRITE ERROR)\n'.format(src.directory,
                                                                           name))
            except PermissionError:
                if not self.ignore_read_errors:
                    raise
                report.write('   ** {}: {} (READ ERROR)\n'.format(src.directory,
                                                                  name))

        def synchronize_files(src, dst):  # pylint: disable=too-many-branches,too-many-statements
            """Synchronize source directory and destination"""
            for name in src.listdir():
                try:
                    src_stat = src.stat(name)
                except (DirectorySyncError, PermissionError, FileNotFoundError):
                    report.write('   ** {}: {} (SRC STAT ERROR)\n'.format(src.directory,
                                                                          name))
                    if not self.ignore_read_errors:
                        raise
                    continue
                dst_stat = None
                # check for symbolic link
                is_src_link = stat.S_ISLNK(src_stat.st_mode)
                if is_src_link and not self.follow_symlinks:
                    continue
                # check for directory
                is_src_dir = stat.S_ISDIR(src_stat.st_mode)
                try:
                    dst_stat = dst.stat(name)
                except (DirectorySyncError, PermissionError):
                    report.write('   ** {}: {} (DST STAT ERROR)\n'.format(src.directory,
                                                                          name))
                    if not self.ignore_write_errors:
                        raise
                    continue
                except FileNotFoundError:
                    if is_src_dir and self.recursive:
                        try:
                            dst.mkdir(name, src_stat.st_mode)
                        except (DirectorySyncError, PermissionError):
                            report.write('   ** {}: {} (CREATE ERROR)\n'.format(src.directory,
                                                                                name))
                            if not self.ignore_write_errors:
                                raise
                            continue

                if is_src_dir:
                    # recursive directory synchronization
                    if self.recursive:
                        try:
                            with src.chdir(name):
                                try:
                                    with dst.chdir(name):
                                        synchronize_files(src, dst)
                                except (DirectorySyncError, PermissionError):
                                    report.write('   ** {}: {} (SYNC ERROR)\n'.format(
                                        src.directory, name))
                                    if not self.ignore_write_errors:
                                        raise
                                    continue
                        except (DirectorySyncError, PermissionError):
                            report.write('   ** {}: {} (READ ERROR)\n'.format(src.directory,
                                                                              name))
                            if not self.ignore_read_errors:
                                raise
                            continue
                else:
                    # check filter
                    if self.filter and not fnmatch.fnmatch(name, self.filter):
                        continue
                    # regular file copy
                    if (dst_stat is None) or \
                            (not self.use_datetime) or \
                            (src_stat.st_mtime > dst_stat.st_mtime) or \
                            (src_stat.st_size != dst_stat.st_size):
                        try:
                            copy_file(src, dst, name, src_stat)
                        except (DirectorySyncError, PermissionError):
                            report.write('   ** {}: {} (WRITE ERROR)\n'.format(src.directory,
                                                                               name))
                            if not self.ignore_write_errors:
                                raise
                            continue
                    else:
                        report.write('    = {}: {} (unchanged)\n'.format(src.directory,
                                                                         name))
                    if self.delete_source:
                        try:
                            src.unlink(name)
                        except (DirectorySyncError, PermissionError):
                            report.write('   ** {}: {} (DELETE ERROR)\n'.format(src.directory,
                                                                                name))
                            if not self.ignore_delete_errors:
                                raise
                            continue

        try:
            report.write('Directories synchronization output\n'
                         '==================================\n'
                         ' - source: {!r}\n'
                         ' - target: {!r}\n'
                         ' - copied files:\n'.format(self.source, self.target))
            with self.source.get_handler() as source:
                with self.target.get_handler() as target:
                    synchronize_files(source, target)
            if copied['files']:
                report.write('   >> {} files ({}) copied\n'.format(
                    len(copied['files']),
                    get_human_size(copied['size'])))
            else:
                report.write('   >> no copied file\n')
            return (TASK_STATUS_OK, copied) if copied['files'] else (TASK_STATUS_EMPTY, None)

        except:  # pylint: disable=bare-except
            etype, value, tb = sys.exc_info()  # pylint: disable=invalid-name
            report.write('\n\n'
                         'An error occurred\n'
                         '=================\n')
            report.write(''.join(traceback.format_exception(etype, value, tb)))
            return TASK_STATUS_FAIL, None



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.scheduler module

This module defines the main persistent scheduler utility class.
"""

__docformat__ = 'restructuredtext'

from zope.container.folder import Folder
from zope.interface import implementer
from zope.intid import IIntIds
from zope.schema.fieldproperty import FieldProperty

from pyams_scheduler.interfaces import IScheduler, ISchedulerHandler, ISchedulerRoles, \
    SCHEDULER_AUTH_KEY, SCHEDULER_HANDLER_KEY
from pyams_security.interfaces import IDefaultProtectionPolicy, IRolesPolicy
from pyams_security.property import RolePrincipalsFieldProperty
from pyams_security.security import ProtectedObjectMixin, ProtectedObjectRoles
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.registry import get_pyramid_registry, query_utility
from pyams_zmq.socket import zmq_response, zmq_socket


@implementer(ISchedulerHandler)
class SchedulerHandler:
    """Scheduler handler utility

    This is just a 'marker' utility which is used to mark nodes in a cluster
    which should run the scheduler
    """


@implementer(IScheduler, IDefaultProtectionPolicy)
class Scheduler(ProtectedObjectMixin, Folder):
    """Scheduler utility"""

    zodb_name = FieldProperty(IScheduler['zodb_name'])
    report_mailer = FieldProperty(IScheduler['report_mailer'])
    report_source = FieldProperty(IScheduler['report_source'])
    notified_host = FieldProperty(IScheduler['notified_host'])
    show_home_menu = FieldProperty(IScheduler['show_home_menu'])

    @property
    def tasks(self):
        """Scheduler tasks getter"""
        yield from self.values()

    @property
    def history(self):
        """Scheduler full history getter"""
        result = []
        for task in self.values():
            result.extend(task.history)
        result.sort(key=lambda x: x.date)
        return result

    @property
    def internal_id(self):
        """Scheduler internal ID"""
        intids = query_utility(IIntIds)
        if intids is not None:
            return intids.register(self)
        return None

    @staticmethod
    def get_socket():
        """Open ØMQ socket"""
        registry = get_pyramid_registry()
        handler = registry.settings.get(SCHEDULER_HANDLER_KEY, False)
        if handler:
            return zmq_socket(handler, auth=registry.settings.get(SCHEDULER_AUTH_KEY))
        return None

    @staticmethod
    def get_task(task_id):
        """Get task matching given ID"""
        intids = query_utility(IIntIds)
        if intids is not None:
            return intids.queryObject(task_id)
        return None

    def get_jobs(self):
        """Getter of scheduler scheduled jobs"""
        socket = self.get_socket()
        if socket is None:
            return [501, "No socket handler defined in configuration file"]
        socket.send_json(['get_jobs', {}])
        return zmq_response(socket)

    def test_process(self):
        """Send test request to scheduler process"""
        socket = self.get_socket()
        if socket is None:
            return [501, "No socket handler defined in configuration file"]
        socket.send_json(['test', {}])
        return zmq_response(socket)


@implementer(ISchedulerRoles)
class SchedulerRoles(ProtectedObjectRoles):
    """Scheduler roles"""

    scheduler_managers = RolePrincipalsFieldProperty(ISchedulerRoles['scheduler_managers'])

    tasks_managers = RolePrincipalsFieldProperty(ISchedulerRoles['tasks_managers'])


@adapter_config(required=IScheduler,
                provides=ISchedulerRoles)
def scheduler_roles_adapter(context):
    """Scheduler roles adapter"""
    return SchedulerRoles(context)


@adapter_config(name='scheduler_roles',
                required=IScheduler,
                provides=IRolesPolicy)
class SchedulerRolesPolicy(ContextAdapter):
    """Scheduler roles policy"""

    roles_interface = ISchedulerRoles
    weight = 10



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.interfaces.task.rest module

This module defines interface of REST API caller task.
"""

from zope.interface import Interface
from zope.schema import Bool, Int, Password, Text, TextLine, URI

from pyams_scheduler.interfaces import ITask
from pyams_utils.schema import HTTPMethodField


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


class IRESTCallerTaskInfo(Interface):
    """REST API caller task info"""

    base_url = URI(title=_("Base target URI"),
                   description=_("Base URI, including protocol and hostname of remote service"),
                   required=True)

    service = HTTPMethodField(title=_("REST service"),
                              description=_("Method and relative URL of REST service"),
                              required=True,
                              default=('GET', '/'))

    params = Text(title=_("Service parameters"),
                  description=_("Enter service parameters, in JSON object format"),
                  required=False)

    verify_ssl = Bool(title=_("Verify SSL?"),
                      description=_("If 'no', SSL certificates will not be verified"),
                      required=False,
                      default=True)

    connection_timeout = Int(title=_("Connection timeout"),
                             description=_("Connection timeout, in seconds; keep empty to use "
                                           "system's default, which is also none by default"),
                             required=False,
                             default=30)

    allow_redirects = Bool(title=_("Allow redirects?"),
                           description=_("If disabled, redirections will now be handled"),
                           required=False,
                           default=True)

    ok_status = TextLine(title=_("OK status"),
                         description=_("Comma-separated list of HTTP status which may be "
                                       "considered as success"),
                         required=True,
                         default='200')

    authenticate = Bool(title=_("Required authentication?"),
                        description=_(""),
                        required=False,
                        default=False)

    username = TextLine(title=_("User name"),
                        description=_("Service login"),
                        required=False)

    password = Password(title=_("Password"),
                        description=_("Service password"),
                        required=False)

    use_proxy = Bool(title=_("Use proxy server?"),
                     description=_("Check if an HTTP proxy is required"),
                     required=False,
                     default=False)

    proxy_server = TextLine(title=_("Proxy server"),
                            description=_("Proxy server name"),
                            required=False)

    proxy_port = Int(title=_("Proxy port"),
                     description=_("Proxy server port"),
                     required=False,
                     default=8080)

    proxy_username = TextLine(title=_("Proxy user name"),
                              required=False)

    proxy_password = Password(title=_("Proxy password"),
                              required=False)

    use_jwt_authority = Bool(title=_("Use JWT authority?"),
                             description=_("If 'yes', get JWT token from authentication "
                                           "authority"),
                             required=False,
                             default=False)

    jwt_authority_url = URI(title=_("JWT authority location"),
                            description=_("Base URL (protocol and hostname) or JWT "
                                          "authentication authority"),
                            required=False)

    jwt_token_service = HTTPMethodField(title=_("Token getter service"),
                                        description=_("Method and relative URL of REST API used "
                                                      "to get access tokens"),
                                        required=False,
                                        default=('POST', '/api/auth/jwt/token'))

    jwt_login_field = TextLine(title=_("JWT login attribute"),
                               description=_("Name of the attribute containing principal ID "
                                             "sent to JWT authority"),
                               required=False,
                               default='login')

    jwt_password_field = TextLine(title=_("JWT password attribute"),
                                  description=_("Name of the attribute containing principal "
                                                "password sent to JWT authority"),
                                  required=False,
                                  default='password')

    jwt_token_attribute = TextLine(title=_("JWT token attribute"),
                                   description=_("Name of the attribute containing the access "
                                                 "token in JSON response"),
                                   required=False,
                                   default='accessToken')

    jwt_use_proxy = Bool(title=_("Use proxy settings for JWT API?"),
                         description=_("If 'yes', proxy settings defined above will also be used "
                                       "to get access to JWT authority"),
                         required=False,
                         default=False)


class IRESTCallerTask(ITask, IRESTCallerTaskInfo):
    """REST API caller interface"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.task.rest module

This module defines REST caller task.
"""

import codecs
import json
import pprint
import sys
import traceback
from urllib import parse

import chardet
import requests
from requests import ConnectionError
from zope.schema.fieldproperty import FieldProperty

from pyams_scheduler.interfaces.task import TASK_STATUS_ERROR, TASK_STATUS_FAIL, TASK_STATUS_OK
from pyams_scheduler.interfaces.task.rest import IRESTCallerTask
from pyams_scheduler.task import Task
from pyams_utils.dict import format_dict
from pyams_utils.factory import factory_config
from pyams_utils.html import html_to_text


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


@factory_config(IRESTCallerTask)
class RESTCallerTask(Task):
    """REST API caller task"""

    label = _("HTTP service")
    icon_class = 'fab fa-battle-net'

    base_url = FieldProperty(IRESTCallerTask['base_url'])
    service = FieldProperty(IRESTCallerTask['service'])
    params = FieldProperty(IRESTCallerTask['params'])
    verify_ssl = FieldProperty(IRESTCallerTask['verify_ssl'])
    connection_timeout = FieldProperty(IRESTCallerTask['connection_timeout'])
    allow_redirects = FieldProperty(IRESTCallerTask['allow_redirects'])
    ok_status = FieldProperty(IRESTCallerTask['ok_status'])
    authenticate = FieldProperty(IRESTCallerTask['authenticate'])
    username = FieldProperty(IRESTCallerTask['username'])
    _password = FieldProperty(IRESTCallerTask['password'])
    use_proxy = FieldProperty(IRESTCallerTask['use_proxy'])
    proxy_server = FieldProperty(IRESTCallerTask['proxy_server'])
    proxy_port = FieldProperty(IRESTCallerTask['proxy_port'])
    proxy_username = FieldProperty(IRESTCallerTask['proxy_username'])
    _proxy_password = FieldProperty(IRESTCallerTask['proxy_password'])
    use_jwt_authority = FieldProperty(IRESTCallerTask['use_jwt_authority'])
    jwt_authority_url = FieldProperty(IRESTCallerTask['jwt_authority_url'])
    jwt_token_service = FieldProperty(IRESTCallerTask['jwt_token_service'])
    jwt_login_field = FieldProperty(IRESTCallerTask['jwt_login_field'])
    jwt_password_field = FieldProperty(IRESTCallerTask['jwt_password_field'])
    jwt_token_attribute = FieldProperty(IRESTCallerTask['jwt_token_attribute'])
    jwt_use_proxy = FieldProperty(IRESTCallerTask['jwt_use_proxy'])

    @property
    def password(self):
        """Password getter"""
        return self._password

    @password.setter
    def password(self, value):
        """Password setter"""
        if value == '*****':
            return
        self._password = value

    @property
    def proxy_password(self):
        """Proxy password getter"""
        return self._proxy_password

    @proxy_password.setter
    def proxy_password(self, value):
        """Proxy password setter"""
        if value == '*****':
            return
        self._proxy_password = value

    @property
    def ok_status_list(self):
        """OK status list getter"""
        return map(int, self.ok_status.split(','))

    def run(self, report, **kwargs):
        # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        # get remote service URL
        method, service = self.service
        rest_service = f'{self.base_url}{service}'
        report.write(f'HTTP service output\n'
                     f'===================\n'
                     f'HTTP service: \n    {method} {rest_service}\n\n')
        # check proxy configuration
        proxies = {}
        if self.use_proxy:
            parsed = parse.urlparse(self.base_url)
            if self.proxy_username:
                proxy_auth = f'{self.proxy_username}:{self.proxy_password}@'
            else:
                proxy_auth = ''
            proxies[parsed.scheme] = f'http://{proxy_auth}{self.proxy_server}:{self.proxy_port}'
        # check JWT authorization
        auth = None
        headers = {}
        if self.use_jwt_authority:
            jwt_method, jwt_service = self.jwt_token_service
            jwt_service = f'{self.jwt_authority_url}{jwt_service}'
            jwt_params = {
                self.jwt_login_field: self.username,
                self.jwt_password_field: self.password
            }
            try:
                jwt_request = requests.request(jwt_method, jwt_service,
                                               headers={
                                                   'Content-Type': 'application/json'
                                               },
                                               params=jwt_params if jwt_method == 'GET' else None,
                                               data=json.dumps(jwt_params)
                                                   if jwt_method != 'GET' else None,
                                               proxies=proxies if self.jwt_use_proxy else None,
                                               timeout=self.connection_timeout,
                                               allow_redirects=False)
            except ConnectionError:
                etype, value, tb = sys.exc_info()  # pylint: disable=invalid-name
                report.write('\n\n'
                             'An HTTP error occurred\n'
                             '======================\n')
                report.write(''.join(traceback.format_exception(etype, value, tb)))
                return TASK_STATUS_FAIL, None
            else:
                status_code = jwt_request.status_code
                report.write(f'JWT token status code: {status_code}\n')
                if status_code != requests.codes.ok:  # pylint: disable=no-member
                    report.write(f'JWT headers: {format_dict(jwt_request.headers)}\n')
                    report.write(f'JWT params: {format_dict(jwt_params)}\n')
                    report.write(f'JWT report: {jwt_request.text}\n\n')
                    return TASK_STATUS_ERROR, None
                headers['Authorization'] = f'Bearer ' \
                                           f'{jwt_request.json().get(self.jwt_token_attribute)}'
        # build authorization headers
        elif self.username:
            auth = self.username, self.password
        # check params
        params = {}
        if self.params:
            params.update(json.loads(self.params))
        params.update(kwargs)
        # build HTTP request
        try:
            rest_request = requests.request(method, rest_service,
                                            auth=auth,
                                            headers=headers,
                                            params=params if method == 'GET' else None,
                                            data=params if method != 'GET' else None,
                                            verify=self.verify_ssl,
                                            proxies=proxies,
                                            timeout=self.connection_timeout,
                                            allow_redirects=self.allow_redirects)
        except ConnectionError:
            etype, value, tb = sys.exc_info()  # pylint: disable=invalid-name
            report.write('\n\n'
                         'An HTTP error occurred\n'
                         '======================\n')
            report.write(''.join(traceback.format_exception(etype, value, tb)))
            return TASK_STATUS_FAIL, None
        else:
            # check request status
            status_code = rest_request.status_code
            report.write(f'Status code: {status_code}\n')
            report.write(f'Headers: {format_dict(rest_request.headers)}\n\n')
            # check request content
            content_type = rest_request.headers.get('Content-Type', 'text/plain')
            if content_type.startswith('application/json'):
                response = rest_request.json()
                message = pprint.pformat(response)
            elif content_type.startswith('text/html'):
                message = html_to_text(rest_request.text)
            elif content_type.startswith('text/'):
                message = rest_request.text
            else:
                content = rest_request.content
                if 'charset=' in content_type.lower():
                    charset = content_type.split('=', 1)[1]
                else:
                    charset = chardet.detect(content).get('encoding') or 'utf-8'
                message = codecs.decode(content, charset)
            report.write(message)
            report.write('\n\n')
            return (
                TASK_STATUS_OK if status_code in self.ok_status_list else status_code,
                rest_request
            )
